//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
unit nparameter;
//==============================================================================
{$DEFINE CPU386}
{$N+ $A8}
//==============================================================================
interface
//==============================================================================

Const
  version_string='lossyWAV 1.1.0c';
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
  lossyWAVHead1=version_string+', Copyright (C) 2007,2008 Nick Currie. Copyleft.'+#13#10;
  lossyWAVhead2= 'This is free software under the GNU GPLv3+ license; There is NO WARRANTY, to'+
          #13#10+'the extent permitted by law. <http://www.gnu.org/licenses/> for details.'+#13#10;

var
  Parameter_String : string[255];

  procedure lossyWAVerror  (lwe_string : string; lwe_value : integer);
  procedure lossyWAVwarning(lww_string : string);

  function  DirInp                : string;   // name of the source file folder.
  function  DirOut                : String;   // name of the output file folder.

  function  wavInp                : string;   // name of the input wav file & directory
  function  wavOut                : string;   // name of the input wav file & directory

  function  lossyInp              : String;   // name of the lossy input wav file.
  function  lossyOut              : string;   // name of the lossy output wav file based on the input file name.

  function  lwcdfInp              : String;   // name of the lwcdf input wav file.
  function  lwcdfOut              : string;   // name of the lwcdf output wav file based on the input file name.

  Function  WAVFilePrintName      : String;   // name of WAV file OR 'from STDIN' if STDINPUT=1.
  function  lossyFileName         : string;   // output filename only;
  function  lwcdfFileName         : string;   // correction file output filename only;
  function  logfilename           : String;   // name of lossyWAV log file.

  function  quality               : Double;   // quality level. default=2;
  function  forcing               : Integer;  // forcibly over-write output file if it exists.
  function  correct               : Integer;  // write correction file if correct=1.
  function  checking              : Integer;

  function  merging               : Integer;
  function  dither                : Double;   // enable TPDF dither when noise shaping.
  function  silent                : Integer;  // enable silent running.

  function  verbosity             : Integer;  // verbosity level. default=false;
  function  warnings              : Integer;  // warnings switch state. default=false;
  function  priority              : Integer;  // below or low_priority process priority switch;

  function  shaping               : Double;   // enable optional fixed noise shaping; default=off.
  function  writetolog            : Integer;  // enable writing to lossyWAV.log file; default=0.

  function  limit                 : Double;   // modify upper bound of frequency range analysed.
  function  scaling               : Double;   // scaling factor from WaveGain;

  function  detailledoutput       : Integer;  // debug switch state. default=false
  function  linkchannels          : Integer;  // revert to "old" method of one bits-to-remove value for all channels.
  function  blockdist             : Integer;  // show bit distribution of WAVE / BTRD output codec blocks;

  function  sampledist            : Integer;  // show bit distribution of WAVE / BTRD output samples;
  function  STDINPUT              : Integer;  // read from STDIN;
  function  STDOUTPUT             : Integer;  // write to STDOUTPUT;
  function  outputname            : String;   // output filename if STDIN input selected.

  procedure open_log_file;
  procedure close_log_file;

//==============================================================================
implementation
//==============================================================================

uses
  SysUtils,
  Windows,
  Math,
  nwav,
  ntypes;

//==============================================================================
//nparameter constants, variables & types
//==============================================================================
const
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
  lossyWAV_GPL = 'This program is free software: you can redistribute it and/or modify it under'+
            crlf+'the terms of the GNU General Public License as published by the Free Software'+
            crlf+'Foundation, either version 3 of the License, or (at your option) any later'+
            crlf+'version.'+crlf+
            crlf+'This program is distributed in the hope that it will be useful,but WITHOUT ANY'+
            crlf+'WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A'+
            crlf+'PARTICULAR PURPOSE.  See the GNU General Public License for more details.'+crlf+
            crlf+'You should have received a copy of the GNU General Public License along with'+
            crlf+'this program.  If not, see <http://www.gnu.org/licenses/>.'+crlf;
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
  lossyWAV_process_description =
                 'Process Description:'+crlf+
            crlf+'lossyWAV adds white noise to the processed output. The amount of added noise is'+
            crlf+'based on analysis of the signal levels in the frequency range 20Hz to 16kHz.'+crlf+
            crlf+'If signals above the upper limiting frequency are at an even lower level, they'+
            crlf+'can be swamped by the added noise. This is usually inaudible, but the behaviour'+
            crlf+'can be changed by specifying a higher --limit (in the range 16kHz to 20kHz).'+crlf+
            crlf+'For many audio signals, there is little content at very high frequencies, and'+
            crlf+'forcing lossyWAV to keep the added noise level lower than the content at these'+
            crlf+'frequencies can increase the bitrate dramatically for no perceptible benefit.'+crlf;
                //12345678901234567890123456789012345678901234567890123456789012345678901234567890
  lossyWAV_standard_help =
                 'Usage   : lossyWAV <input wav file> <options>'+
            crlf+//345678901234567890123456789012345678901234567890123456789012345678901234567890
            crlf+'Example : lossyWAV musicfile.wav'+
            crlf+//345678901234567890123456789012345678901234567890123456789012345678901234567890
            crlf+'Quality Options:'+
            crlf+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            crlf+'-I, --insane        highest quality output, suitable for transcoding;'+
            crlf+'-E, --extreme       high quality output, also suitable for transcoding;'+
            crlf+'-S, --standard      default quality output, considered to be transparent;'+
            crlf+'-P, --portable      good quality output for DAP use. Not considered to be fully'+
            crlf+'                    transparent, but considered fit for its intended purpose.'+
            crlf+
            crlf+'Standard Options:'+
            crlf+//345678901234567890123456789012345678901234567890123456789012345678901234567890
            crlf+'-c, --check         check if WAV file has already been processed; default=off.'+
            crlf+'                    errorlevel=16 if already processed, 0 if not.'+
            crlf+'-C, --correction    write correction file for processed WAV file; default=off.'+
            crlf+'-f, --force         forcibly over-write output file if it exists; default=off.'+
            crlf+'-h, --help          display help.'+
            crlf+'-L, --longhelp      display extended help.'+
            crlf+'-M, --merge         merge existing lossy.wav and lwcdf.wav files.'+
            crlf+'-o, --outdir <t>    destination directory for the output file(s).'+
            crlf+'-v, --version       display the lossyWAV version number.'+crlf;
  lossyWAV_advanced_help =
            crlf+'Advanced Options:'+
            crlf+//345678901234567890123456789012345678901234567890123456789012345678901234567890
            crlf+'-                   if filename="-" then WAV input is taken from STDIN.'+
            crlf+'    --blockdist     show distribution of lowest significant bit of input'+
            crlf+'                    codec-blocks and bit-removed codec-blocks.'+
            crlf+'-D, --dither <n>    enable variable PDF dither of output; default=off; '+
            crlf+'                    0 = rectangular; 1 = triangular; 0.5 = half way between.'+
            crlf+'-l, --limit <n>     set upper frequency limit to be used in analyses to n Hz;'+
            crlf+'                    (16000<=n<=20000), default = 16000.'+
            crlf+'    --linkchannels  Revert to original single bits-to-remove value for all'+
            crlf+'                    channels rather than channel dependent bits-to-remove.'+
            crlf+'-q, --quality <n>   quality preset (10=highest quality, 0=lowest bitrate;'+
            crlf+'                    default = --standard = 5; --insane = 10; --extreme = 7.5;'+
            crlf+'                    --portable = 2.5)'+
            crlf+'    --sampledist    show distribution of lowest significant bit of input'+
            crlf+'                    samples and bit-removed samples.'+
            crlf+'    --scale <n>     scaling factor from WaveGain, etc; (0.0<n<=8.0),default=1.0'+
            crlf+'-s, --shaping <n>   enable fixed noise shaping; (0.00<=n<=1.00); default=q/10;'+
            crlf+'                    0.00 = off, 1.00 = 100% effectiveness, 0.50 = 50%, etc.'+
            crlf+'    --stdinname <t> pseudo filename to use when input from STDIN.'+
            crlf+'    --stdout        write processed WAV output to STDOUT.'+
            crlf+'-w, --writetolog    create (or append to) lossyWAV.log in the output directory.'+
            crlf+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            crlf+'System Options:'+
            crlf+//2345678901234567890123456789012345678901234567890123456789012345678901234567890
            crlf+'-B, --below         set process priority to below normal.'+
            crlf+'-d, --detail        enable detailed bits-to-remove information output mode'+
            crlf+'    --low           set process priority to low.'+
            crlf+'-n, --nowarnings    suppress lossyWAV warnings.'+
            crlf+'-Q, --quiet         significantly reduce screen output.'+
            crlf+'    --silent        no screen output.'+
            crlf;//345678901234567890123456789012345678901234567890123456789012345678901234567890
  lossyWAV_special_thanks =
            crlf+'Special thanks:'+crlf+
            crlf+'David Robinson      for the publication of his lossyFLAC method, guidance, and'+
            crlf+'                    the motivation to implement the method as lossyWAV.'+
            crlf+'Horst Albrecht      for ABX testing, valuable support in tuning the internal'+
            crlf+'                    presets, constructive criticism and all the feedback.'+
            crlf+'Sebastian Gesemann  for the noise shaping coefficients and help in using them'+
            crlf+'                    in the lossyWAV noise shaping implementation.'+
            crlf+'Don Cross           for the Complex-FFT algorithm used.'+
            crlf;//345678901234567890123456789012345678901234567890123456789012345678901234567890
var
    parmError                 : string;
    WavInpDir                 : string;
    WavOutDir                 : string;
    logfilename_int           : String;

    wavName                   : String;
    pseudoname                : String;
    lossyName                 : String;
    lwcdfName                 : String;

    quality_int               : Double;
    scaling_int               : Double;

    shaping_int               : Double;
    highskew_int              : Double;

    minbits_int               : Double;
    dither_int                : Double;

    noclips_int               : Integer;
    detail_int                : Integer;
    analyses_int              : Integer;
    fft32_int                 : Integer;

    verbosity_int             : Integer;
    warnings_int              : Integer;
    forcing_int               : Integer;
    NextParamStrParamNr       : Integer; // internal param counter for nextParamStr

    priority_int              : Integer;
    correct_int               : Integer;
    checking_int              : Integer;
    merging_int               : Integer;

    help_int                  : Integer;
    silent_int                : Integer;
    linkchannels_int          : Integer;
    writetolog_int            : Integer;

    STDINPUT_int              : Integer;
    STDOUTPUT_int             : Integer;
    blocksize_int             : Integer;
    blockdist_int             : Integer;

    sampledist_int            : Integer;
    limit_int                 : Integer;
    logfileopened_int         : Integer;
    dontusegetlasterror_int   : Integer;

    nparameter_packer         : array[0..3] of integer;

  procedure lossyWAVerror(lwe_string : string; lwe_value : integer);
  begin
    if (silent=0) then
      write(ErrOutput,'%lossyWAV Error% : ',lwe_string,crlf);
    halt(lwe_value);
  end;

  procedure lossyWAVwarning(lww_string : string);
  begin
    if (warnings=1) and (silent=0) then
      write(ErrOutput,'%lossyWAV Warning% : ',lww_string,crlf);
  end;

  procedure parmsError;
  begin
    if (paramcount>0) and (help_int=0) then
      write(ErrOutput,parmerror,crlf)
    else
    begin
      if help_int>1 then
        write(ErrOutput,lossyWAV_process_description+crlf);
      write(ErrOutput,lossyWAV_standard_help);
      if help_int>1 then
        write(ErrOutput,lossyWAV_advanced_help);
      write(ErrOutput,lossyWAV_Special_Thanks);
    end;
  end;

  procedure initNextParamStr;
  begin
   NextParamStrParamNr:=0;
  end;

  function nextParamStr(var parm: string): Integer;
  begin
    if NextParamStrParamNr>=paramCount then
    begin
      nextParamStr:=0;
      exit;
    end;
    NextParamStrParamNr:=NextParamStrParamNr + 1;
    parm:=ParamStr(NextParamStrParamNr);
    nextParamStr:=1;
  end;

  procedure parmerror_multiple_selection;
  begin
    lossyWAVerror('Multiple '+parmerror+' switches given.',$31);
  end;

  procedure parmerror_mutually_incompatible_selection;
  begin
    lossyWAVerror('Switches '+parmerror+' are incompatible.',$31);
  end;

  procedure parmerror_no_value_given;
  begin
    lossyWAVerror('No '+parmerror+' value given.',$31);
  end;

  procedure parmerror_val_error;
  begin
    lossyWAVerror('Error evaluating '+parmerror+' value given.',$31);
  end;

  procedure writeversionstring;
  begin
    write(ErrOutput,version_string,crlf);
    halt(0);
  end;

  procedure check_permitted_values(cpv_var, cpv_low, cpv_high : Integer); Overload;
  begin
    if (cpv_var<cpv_low) or (cpv_var>cpv_high) then
      lossyWAVerror('Permitted '+parmerror+' values : '+inttostr(cpv_low)+'<=n<='+inttostr(cpv_high),$31);
  end;

  procedure check_permitted_values(cpv_var, cpv_low, cpv_high : Double;digits : Integer); Overload;
  begin
    if (cpv_var<cpv_low) or (cpv_var>cpv_high) then
      lossyWAVerror('Permitted '+parmerror+' values : '+floattostrf(cpv_low,ffFixed,99,digits)+'<=n<='+floattostrf(cpv_high,ffFixed,99,digits),$31);
  end;

  function check_quality(s : string) : boolean;
  var
    valError : Integer;
  begin
    check_quality:=false;

    if (s='-v') or (s='--version') then
      writeversionstring;

    if (s='-h') or (s='--help') then
    begin
      write(ErrOutput,lossyWAVhead1,crlf);
      write(ErrOutput,lossyWAV_GPL,crlf);
      help_int:=1;
      parmserror;
      halt(0);
    end;

    if (s='-L') or (s='--longhelp') then
    begin
      write(ErrOutput,lossyWAVhead1,crlf);
      write(ErrOutput,lossyWAV_GPL,crlf);
      help_int:=2;
      parmserror;
      halt(0);
    end;

    if (s='-q') or (s='--quality') then
    begin
      parmerror:='quality preset';

      if quality_int<>-1 then
        parmerror_multiple_selection;

      if nextParamStr(s)=0 then
        parmerror_no_value_given;

      Val(s, quality_Int, valError);
      if valError <> 0 then
        parmerror_val_error;

      quality_int:=abs(round(quality_int*10000)/10000);

      check_permitted_values(quality_int,0,10,4);

      if quality_int=int(quality_int) then
        parameter_string:='-q '+inttostr(floor(quality_int))
      else
        parameter_string:='-q '+floattostrf(quality_int,ffFixed,99,4);

      check_quality:=true;
    end;

    if (s='-I') or (s='--insane') then
    begin
      parmerror:='quality preset';
      if quality_int<>-1 then
        parmerror_multiple_selection;
      quality_int:=10;
      parameter_string:='--insane';
      check_quality:=true;
    end;

    if (s='-E') or (s='--extreme') then
    begin
      parmerror:='quality preset';
      if quality_int<>-1 then
        parmerror_multiple_selection;
      quality_int:=7.5;
      parameter_string:='--extreme';
      check_quality:=true;
    end;

    if (s='-S') or (s='--standard') then
    begin
      parmerror:='quality preset';
      if quality_int<>-1 then
        parmerror_multiple_selection;
      quality_int:=5;
      parameter_string:='--standard';
      check_quality:=true;
    end;

    if (s='-P') or (s='--portable') then
    begin
      parmerror:='quality preset';
      if quality_int<>-1 then
        parmerror_multiple_selection;
      quality_int:=2.5;
      parameter_string:='--portable';
      check_quality:=true;
    end;
  end;

  function check_system(s : string) : boolean;
  begin
    check_system:=false;

    if             (s='--stdout') then
    begin
      parmerror:='output to STDOUT';
      if STDOUTPUT_int=1 then
        parmerror_multiple_selection;
      STDOUTPUT_int:=1;
      check_system:=true;
    end;

    if (s='-B') or (s='--below') then
    begin
      parmerror:='process priority';
      if (priority_int>0) then
        parmerror_multiple_selection;
      priority_int:=1;
      check_system:=true;
    end;

    if             (s='--low') then
    begin
      parmerror:='process priority';
      if (priority_int>0) then
        parmerror_multiple_selection;
      priority_int:=2;
      check_system:=true;
    end;

  end;

  function check_output(s : string) : boolean;
  begin
    check_output:=false;

    if (s='-d') or (s='--detail') then
    begin
      parmerror:='detail';
      if (detail_int=1) then
        parmerror_multiple_selection;
      detail_int:=1;
      check_output:=true;
    end;

    if             (s='--blockdist') then
    begin
      parmerror:='block lsb distribution';
      if (blockdist_int=1) then
        parmerror_multiple_selection;
      blockdist_int:=1;
      check_output:=true;
    end;

    if             (s='--sampledist') then
    begin
      parmerror:='sample lsb distribution';
      if (sampledist_int=1) then
        parmerror_multiple_selection;
      sampledist_int:=1;
      check_output:=true;
    end;

    if (s='-Q') or (s='--quiet') then
    begin
      parmerror:='reduced output';
      if verbosity_int=0 then
        parmerror_multiple_selection;
      verbosity_int:=0;
      check_output:=true;
    end;

    if             (s='--silent') then
    begin
      parmerror:='silent running';
      if silent_int=1 then
        parmerror_multiple_selection;
      silent_int:=1;
      check_output:=true;
    end;

    if (s='-n') or (s='--nowarnings') then
    begin
      parmerror:='no warnings';
      if warnings_int=0 then
        parmerror_multiple_selection;
      warnings_int:=0;
      check_output:=true;
    end;

    if (s='-c') or (s='--check') then
    begin
      parmerror:='check';
      if checking_int=1 then
        parmerror_multiple_selection;
      checking_int:=1;
      check_output:=true;
    end;

    if (s='-M') or (s='--merge') then
    begin
      parmerror:='merge';
      if merging_int=1 then
        parmerror_multiple_selection;
      merging_int:=1;
      check_output:=true;
    end;
  end;

  function check_file_output(s : string) : boolean;
  begin
    check_file_output:=false;

    if (s='-w') or (s='--writetolog') then
    begin
      parmerror:='write log output';
      if writetolog_int=1 then
        parmerror_multiple_selection;
      writetolog_int:=1;
      check_file_output:=true;
    end;

    if             (s='--stdinname') then
    begin
      parmerror:='pseudo filename';
      if pseudoname<>'' then
        parmerror_multiple_selection;
      if nextParamStr(s)=0 then
        parmerror_no_value_given;
      pseudoname:=s;
      check_file_output:=true;
    end;

    if (s='-f') or (s='--force') then
    begin
      parmerror:='forced over-write';
      if forcing_int=1 then
        parmerror_multiple_selection;
      forcing_int:=1;
      check_file_output:=true;
    end;

    if (s='-o') or (s='--outdir') then
    begin
      parmerror:='destination folder';
      if (WAVOutDir <> '') then
        parmerror_multiple_selection;
      if nextParamStr(s)=0 then
        parmerror_no_value_given;
      WAVOutDir:=s;
      if not((WAVOutDir[length(WAVOutDir)]='\') or (WAVOutDir[length(WAVOutDir)]=':')) then
        WAVOutdir:=WAVOutdir+'\';
      check_file_output:=true;
    end;

    if (s='-C') or (s='--correction') then
    begin
      parmerror:='correction file';
      if correct_int=1 then
        parmerror_multiple_selection;
      correct_int:=1;
      check_file_output:=true;
    end;
  end;

  function check_advanced(s : string) : boolean;
  var
    valError : Integer;
  begin
    check_advanced:=false;

    if (s='-s') or (s='--shaping') then
    begin
      parmerror:='noise shaping';
      if shaping_int<>-1 then
        parmerror_multiple_selection;

      if nextParamStr(s)=0 then
        parmerror_no_value_given;
      Val(s, shaping_int, valError);
      if valError <> 0 then
        parmerror_val_error;
      shaping_int:=round(shaping_int*1000)/1000;
      check_permitted_values(shaping_int,0.0,1.0,3);
      check_advanced:=true;
    end;

    if (s='-D') or (s='--dither') then
    begin
      parmerror:='dither';
      if dither_int>-1 then
        parmerror_multiple_selection;
      if nextParamStr(s)=0 then
        parmerror_no_value_given;
      Val(s, dither_int, valError);
      if valError <> 0 then
        parmerror_val_error;
      dither_int:=round(dither_int*10000)/10000;
      check_permitted_values(dither_int,0.0,1.0,4);
      check_advanced:=true;
    end;

    if             (s='--scale') then
    begin
      parmerror:='scaling factor';
      if scaling_int<>-1 then
        parmerror_multiple_selection;
      if nextParamStr(s)=0 then
        parmerror_no_value_given;
      Val(s, scaling_int, valError);
      if valError <> 0 then
        parmerror_val_error;
      scaling_int:=round(scaling_int*1000000)/1000000;
      check_permitted_values(scaling_int,0.000001,8.0,6);
      check_advanced:=true;
    end;

    if (s='-l') or (s='--limit') then
    begin
      parmerror:='limit frequency';
      if limit_int<>-1 then
        parmerror_multiple_selection;
      if nextParamStr(s)=0 then
        parmerror_no_value_given;
      Val(s, limit_int, valError);
      if valError <> 0 then
        parmerror_val_error;
      check_permitted_values(limit_int,16000,24000);
      check_advanced:=true;
    end;

    if             (s='--linkchannels') then
    begin
      parmerror:='link channels';
      if linkchannels_int<>0 then
        parmerror_multiple_selection;
      linkchannels_int:=1;
      check_advanced:=true;
    end;
  end;

  function getParms: Integer;
  var
    s        : string;
    colonpos : Integer;
    slashpos : Integer;
  begin
    getParms:=0;
    help_int:=0;
    minbits_int:=-1;
    forcing_int:=0;
    fft32_int:=0;
    correct_int:=0;
    checking_int:=0;
    merging_int:=0;
    noclips_int:=0;
    analyses_int:=0;
    linkchannels_int:=0;
    verbosity_int:=1;
    limit_int:=-1;
    warnings_int:=1;
    shaping_int:=-1;
    scaling_int:=-1;
    highskew_int:=-1;
    quality_int:=-1;
    priority_int:=0;
    detail_int:=0;
    dither_int:=-1;
    silent_int:=0;
    WAVOutDir:='';
    wavName:='';
    writetolog_int:=0;
    blocksize_int:=0;
    blockdist_int:=0;
    dontusegetlasterror_int:=0;

    initNextParamStr;

    if nextParamStr(s)=0 then
      exit;

    if (s='-') then
    begin
      STDINPUT_int:=1;
      wavname:='';
      wavinpdir:='';
    end
    else
      if (s[1]='-') then
      begin
        if (s='-v') or (s='--version') then
          writeversionstring
        else
        begin
          write(ErrOutput,lossyWAVhead1,crlf);
          write(ErrOutput,lossyWAV_GPL,crlf);
          if (s='-h') or (s='--help') or (s='-L') or (s='--longhelp') then
          begin
            help_int:=1+ord((s='--longhelp') or (s='-L'));
            parmserror;
            halt(0);
          end
          else
            lossyWAVerror('No input file given.',1);
        end;
      end
      else
      begin
        wavName:=s;
        wavInpDir:='';
        STDINPUT_int:=0;
      end;

    if (wavname<>'') then
    begin
      if (wavName[1]='"') then
        wavName:=copy(wavName,2,length(wavName)-1);
      if (wavName[length(wavName)]='"') then
        wavName:=copy(wavName,1,length(wavName)-1);

      colonpos:=pos(':',wavName);
      if colonpos>0 then
      begin
        wavInpDir:=copy(wavName,1,colonpos);
        wavName:=copy(wavName,colonpos+1,length(wavName)-colonpos);
      end;

      slashpos:=pos('\',wavName);
      while slashpos>0 do
      begin
        wavInpDir:=wavInpDir+copy(wavName,1,slashpos);
        wavName:=copy(wavName,slashpos+1,length(wavName)-slashpos);
        slashpos:=pos('\',wavName);
      end;
    end;

    while nextParamStr(s)=1 do
    begin
      if not check_quality(s) then
        if not check_system(s) then
          if not check_output(s) then
            if not check_file_output(s) then
              if not check_advanced(s) then
              begin
                parmError:='%lossyWAV Error% : Incorrect option: "' + s + '"'+crlf;
                exit;
              end;
    end;
    getParms:=1;
  end;

  function DirInp : String;
  begin
    DirInp:=wavInpDir;
  end;

  function DirOut : String;
  begin
    DirOut:=wavOutDir;
  end;

  Function wavFilePrintName : String;
  Begin
    if STDINPUT=1 then
    begin
      if pseudoname='' then
        wavFilePrintName:='from STDIN'
      else
        wavfileprintname:=pseudoname;
    end
    else
      wavFilePrintName:=wavname;
  End;

  Function lossyFileName : String;
  Begin
    lossyFileName:=lossyName;
  End;

  Function lwcdfFileName : String;
  Begin
    lwcdfFileName:=lwcdfName;
  End;

  function wavInp: string;
  begin
    wavInp:=wavInpDir + wavName;
  end;

  Function WAVOut: String;
  Begin
    WAVOut:=WAVOutDir + WAVName;
  End;

  function lossyInp: String;
  begin
    lossyInp:=WAVInpDir + lossyFileName;
  end;

  function lossyOut: string;
  begin
    lossyOut:=WAVOutDir + lossyFileName;
  end;

  function lwcdfInp: String;
  begin
    lwcdfInp:=WAVInpDir + lwcdfFileName;
  end;

  function lwcdfOut: string;
  begin
    lwcdfOut:=WAVOutDir + lwcdfFileName;
  end;

  function priority : Integer;
  begin
    priority:=priority_int;
  end;

  function quality: Double;
  begin
    quality:=quality_int;
  end;

  function forcing : Integer;
  begin
    forcing:=forcing_int;
  end;

  function fft32 : Integer;
  begin
    fft32:=fft32_int;
  end;

  function correct : Integer;
  begin
    correct:=correct_int;
  end;

  function checking : Integer;
  begin
    checking:=checking_int;
  end;

  function merging : Integer;
  begin
    merging:=merging_int;
  end;

  function dither : Double;
  begin
    dither:=dither_int;
  end;

  function silent : Integer;
  begin
    silent:=silent_int;
  end;

  function verbosity : Integer;
  begin
    verbosity:=verbosity_int;
  end;

  function detailledoutput : Integer;
  begin
    detailledoutput:=detail_int;
  end;

  function warnings : Integer;
  begin
    warnings:=warnings_int;
  end;

  function shaping : Double;
  begin
    shaping:=shaping_int;
  end;

  function writetolog : Integer;
  Begin
    writetolog:=writetolog_int;
  End;

  function linkchannels : Integer;
  begin
    linkchannels:=linkchannels_int;
  end;

  function STDINPUT : Integer;
  begin
    STDINPUT:=STDINPUT_int;
  end;

  function STDOUTPUT : Integer;
  begin
    STDOUTPUT:=STDOUTPUT_int;
  end;

  function outputname : String;
  begin
    outputname:=pseudoname;
  end;

  function highskew : Double;
  begin
    highskew:=highskew_int;
  end;

  function limit : Double;
  begin
    limit:=limit_int;
  end;

  function blockdist : Integer;
  begin
    blockdist:=blockdist_int;
  end;

  function sampledist : Integer;
  begin
    sampledist:=sampledist_int;
  end;

  function scaling : Double;
  begin
    scaling:=scaling_int;
  end;

  function logfilename : String;
  begin
    logfilename:=logfilename_int;
  end;
  {$I-}
  procedure open_log_file;
  var
    open_retries : Integer;
    lastioresult : Integer;
  begin
    Open_Retries:=0;
    if logfileopened_int=0 then
    begin
      assign(ToLogFile,logfilename_int);
      if fileexists(logfilename_int) then
      begin
        if filegetattr(logfilename_int) and 1=1 then
        Begin
          lossyWAVwarning('lossyWAV.log file is read-only.');
          if(filesetattr(logfilename_int,filegetattr(logfilename_int) and $FE)<>0) then
           lossyWAVerror('Cannot gain write access to '+logfilename_int,$31);
        End;
        Repeat
          {$I-}
          Append(ToLogFile);
          lastIOResult:=ioresult;
          {$I+}
          if lastioresult<>0 then
          begin
            open_retries:=open_retries+1;
            ndelay(powersoftwoinv[6]);
          end;
        Until (lastioresult=0) or (open_retries=128);
        if lastioresult<>0 then
          lossywaverror('Opening lossyWAV.log',$31);
//        if open_retries>0 then
//          write(ToLogFile,'Retries: ',open_retries,crlf);
      end
      else
      begin
        Rewrite(ToLogFile);
        if ioresult<>0 then
          lossywaverror('Opening lossyWAV.log',$31);
      end;
    end;
    logfileopened_int:=1;
  end;

  procedure close_log_file;
  begin
    if logfileopened_int=1 then
    begin
      flush(tologfile);
      close(tologfile);
    end;
    logfileopened_int:=0;
  end;
  {$I+}
//==============================================================================
initialization
//==============================================================================
  logfileopened_int:=0;
  {$I-}
  if (getparms=0) then
  begin
    write(ErrOutput,lossyWAVhead1,crlf);
    write(ErrOutput,lossyWAV_GPL,crlf);
    if parmerror<>'' then
      write(ErrOutput,parmerror);
    write(ErrOutput,'Syntax: lossyWAV <wavfilename>'+crlf);
    write(ErrOutput,crlf+'Type ''lossyWAV --help'' for more detail.'+crlf);
    halt(1);
  end;

  if (help_int>=1) then
  begin
    write(ErrOutput,lossyWAVhead1,crlf);
    write(ErrOutput,lossyWAV_GPL,crlf);
    parmserror;
    halt(1);
  end;

  if (verbosity=1) and (silent=0) then
    write(ErrOutput,lossyWAVhead1,lossywavhead2);

  if (STDINPUT=0) and (length(wavName)=0) then
    lossyWAVerror('Name of input or output wav file missing.',$31);

  if (STDINPUT=1) then
  begin
    wavname:='';
    lossyName:='lossywav.lossy.wav';
    lwcdfname:='lossywav.lwcdf.wav';
    if (merging_int>0) then
      lossyWAVerror('Merge parameter is incompatible'+Crlf+'                   with STDIN file input mode.',$31);
  end
  else
  begin
    if not fileexists(wavInp) then
    begin
      if not fileexists(wavInp+'.wav') then
        lossyWAVerror('Input file: '+wavName+' does not exist.',$31)
      else
        wavName:=wavName+'.wav';
    end;
    lossyName:=copy(wavName, 1, length(wavName) - 3) + 'lossy.' + copy(wavName, length(wavName) - 2, 3);
    lwcdfName:=copy(wavName, 1, length(wavName) - 3) + 'lwcdf.' + copy(wavName, length(wavName) - 2, 3);
  end;

  if (STDOUTPUT=1) then
  begin
    lossyname:='';
    lwcdfname:='';
    wavoutdir:='';
    if (correct_int>0) then
      lossyWAVerror('Correct parameter is incompatible'+Crlf+'                   with STDOUT file output mode.',$31);
  end
  else
  begin
    if (WAVOutdir<>'') and (directoryexists(WAVOutdir)=false) then
      lossyWAVerror('Directory : '+WAVOutdir+' cannot be accessed.',$31);
  end;

  if merging_int=0 then
  begin
    if (STDOUTPUT=0) and (fileexists(lossyOut)) then
    begin
      if forcing_int=1 then
      Begin
        if filegetattr(lossyOut) and 1=1 then
        Begin
          lossyWAVwarning('Output lossy file is read-only.');
          if(filesetattr(lossyOut,filegetattr(lossyOut) and $FE)<>0) then
            lossyWAVerror('Cannot gain write access to '+lossyOut,$31);
        End;
      End
      else
        lossyWAVerror('Output file: '+lossyOut+' exists.',$31);
    end;

    if (correct=1) and (fileexists(lwcdfOut)) then
    begin
      if (forcing=1) then
      Begin
        if filegetattr(lwcdfOut) and 1=1 then
        Begin
          lossyWAVwarning('Output correction file is read-only.');
          if(filesetattr(lwcdfOut,filegetattr(lwcdfOut) and $FE)<>0) then
            lossyWAVerror('Cannot gain write access to '+lwcdfOut,$31);
        End;
      End
      else
        lossyWAVerror('Output file: '+lwcdfOut+' exists.',$31);
    end;
  end
  else
  begin
    lossyName:=wavName;
    if (copy(lossyName,length(lossyName)-9,10)='.lossy.wav') then
    begin
      lwcdfName:=copy(wavName,1,length(wavName)-10)+'.lwcdf.wav';
      wavName:=copy(wavName,1,length(wavName)-10)+'.wav';
    end;

    if fileexists(WAVOut) and (forcing=0) then
      lossyWAVerror('Output merged file already exists',$31);

    if not fileexists(lossyInp) then
      lossyWAVerror('Input lossy file not found.',$31);

    if not fileexists(lwcdfInp) then
      lossyWAVerror('Input correction file not found.',$31);
  end;

  logfilename_int:=DirOut+'lossyWAV.log';

  {$I+}
  nparameter_packer[0]:=0;

  hStdIn :=GetStdHandle(STD_INPUT_HANDLE);
  hStdOut:=GetStdHandle(STD_OUTPUT_HANDLE);
  hStdErr:=GetStdHandle(STD_ERROR_HANDLE);

  ftSTDOUT:=GetFileType(hStdIn);
  ftSTDOUT:=GetFileType(hStdOut);
  ftSTDERR:=GetFileType(hStdErr);
//==============================================================================
end.
//==============================================================================

